/*
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 *
 * The Original Code is ForumZilla 0.1.1.
 *
 * The Initial Developer of the Original Code is Zapogee, Inc.
 * Portions created by Zapogee, Inc. are Copyright (C) 2001
 * Zapogee, Inc.  All Rights Reserved.
 *
 * Contributor(s):
 *   Myk Melez, <myk@melez.com>, original author
 */


//****************************************************************************//
// constants and global variables


//****************************************************************************//
// utility functions


//****************************************************************************//
// nsFeedService component

function FeedService() {}

const MODULE_NAME         = "nsFeedService";
const MODULE_CONTRACT_ID  = "@forumzilla.mozdev.org/feed/feedservice;1";
const MODULE_CLASS_ID     = "{5015eeed-7189-4ddc-984d-c87b5d948759}";
const MODULE_CONSTRUCTOR  = FeedService;

FeedService.interfaces = [Components.interfaces.nsIFeedService,
                          Components.interfaces.nsIObserver,
                          Components.interfaces.nsISupports];

FeedService.prototype.QueryInterface = function (iid) {
    for ( var i=0 ; i<FeedService.interfaces.length ; i++) {
        if (iid.equals(FeedService.interfaces[i]))
            return this;
    }
    Components.returnCode = Components.results.NS_ERROR_NO_INTERFACE;
    return null;
}


//****************************************************************************//
// private methods


//****************************************************************************//
// nsIObserver interface

FeedService.prototype.observe = function (aSubject, aTopic, aData) {
    // Flushes data sources when the user shuts down or changes profiles
    // since autoflushing fails when XPCOM components like this one hold
    // references to the data sources to be flushed.  For more info, see:
    // http://bonsai.mozilla.org/cvsblame.cgi?file=mozilla/rdf/base/src/nsRDFXMLDataSource.cpp&mark=795-798#788

    // No need to check the topic and data arguments, since we only observe
    // the "profile-before-change" topic and will only ever see its
    // "shutdown-persist" context (which is what gets passed as data).

    this.flushDataSource(this.subscriptionsDS);
    this.flushDataSource(this.itemsDS);
}


//****************************************************************************//
// nsIFeedService interface

var gMailWindow;
FeedService.prototype.__defineGetter__('mailWindow', function() {
    if (!gMailWindow) {
        var windowMediator =
          Components
            .classes['@mozilla.org/appshell/window-mediator;1']
              .getService(Components.interfaces.nsIWindowMediator);
        gMailWindow = windowMediator.getMostRecentWindow("mail:3pane");
    }
    return gMailWindow;
});

FeedService.prototype.__defineGetter__('pollingInterval', function() {
    return nsPreferences.getIntPref("forumzilla.polling.interval", DEFAULT_POLLING_INTERVAL);
});

FeedService.prototype.__defineGetter__('storageMode', function() {
    return nsPreferences.getIntPref("forumzilla.storage.mode", DEFAULT_STORAGE_MODE);
});

FeedService.prototype.__defineGetter__('storageFolder', function() {
    return nsPreferences.copyUnicharPref("forumzilla.storage.folder", DEFAULT_STORAGE_FOLDER);
});

// Download a feed.
FeedService.prototype.downloadFeed = function(resource) {
    this.getFeed(resource).download(true, true);
}

// Cache feeds so we don't keep creating them over and over and can hang
// XMLHttpRequest objects off them that they can retrieve after the request
// is complete.
FeedService.feeds = new Object();

FeedService.prototype.getFeed = function(resource) {
    resource = resource.QueryInterface(Components.interfaces.nsIRDFResource);
    var id = resource.Value;

    if (id in FeedService.feeds) {
        var feed = FeedService.feeds[id];
        //debug(id + " requested from feed service, and found in cache");
    }
    else {
        //debug(id + " requested from feed service, but not in cache; creating");
        feed =
          Components
            .classes["@forumzilla.mozdev.org/feed/feed;1"]
              .createInstance(Components.interfaces.nsIFeed);
        feed.init(resource);
        FeedService.feeds[id] = feed;
    }

    return feed;
}

// Download feeds and check for new messages.
FeedService.prototype.downloadFeeds = function() {
    debug("polling feeds");

    var ds = this.subscriptionsDS;
    var feeds = this.subscriptionsList.GetElements();
  
    while(feeds.hasMoreElements()) {
        var resource = feeds.getNext();
        var feed = this.getFeed(resource);

        // Don't poll the feed if it has been disabled.
        if (!feed.enabled) {
            debug(feed.url + " not downloaded; disabled");
            continue;
        }

        // Don't poll the feed if it isn't time yet.
        var elapsed = (new Date().getTime() - feed.lastPolled)/1000/60;
        if (elapsed < feed.interval) {
            debug(feed.url + " not downloaded; " + elapsed.toPrecision(2) + "/" + feed.interval);
            continue;
        }

        feed.download(true, true);
    }
}


FeedService.prototype.flushDataSources = function() {
    
    var ds;

    try {
        ds = this.subscriptionsDS;
        ds = ds.QueryInterface(Components.interfaces.nsIRDFRemoteDataSource);
        if (ds)
            ds.Flush();
    }
    catch(e) {
        debug("couldn't flush subscriptions data source: " + e);
    }

    try {
        ds = this.itemsDS;
        ds = ds.QueryInterface(Components.interfaces.nsIRDFRemoteDataSource);
        if (ds)
            ds.Flush();
    }
    catch(e) {
        debug("couldn't flush items data source: " + e);
    }
}

FeedService.prototype.flushDataSource = function(ds) {
    try {
        ds = ds.QueryInterface(Components.interfaces.nsIRDFRemoteDataSource);
        if (ds)
            ds.Flush();
        else
            throw("data source not defined");
    }
    catch(e) {
        debug("couldn't flush subscriptions data source: " + e);
    }
}

FeedService.prototype.createFolder = function(uri) {
    debug("trying to create folder " + uri);
    var folder;
    try {
        folder = rdf.GetResource(uri).QueryInterface(Components.interfaces.nsIMsgFolder);
        var urlListener = {
            OnStartRunningUrl: function(url) {},
            OnStopRunningUrl: function(url, exitCode) {}
        };
        folder.createStorageIfMissing(urlListener);
    }
    catch(e) {
        debug("error creating folder: " + e);
    }
    return folder;
}

FeedService.prototype.createDefaultFolder = function(title) {
    debug("trying to create default folder");

    var uri;

    if (this.storageMode == FOLDER_STORAGE_MODE)
        uri = this.storageFolder;

    else if (this.storageMode == SUBFOLDERS_STORAGE_MODE) {
        debug("creating default folder as child of " + this.storageFolder);
        var parent = this.mailWindow.GetMsgFolderFromUri(this.storageFolder, true);
        if (!parent) {
            debug("parent folder does not exist; creating it");
            // This could be async, in which case the rest of the function
            // may fail.  XXX We should account for this and make all folder
            // creation be asynchronous.
            parent = this.createFolder(this.storageFolder);
        }
        if (!parent.containsChildNamed(title))
            parent.createSubfolder(title, this.mailWindow.msgWindow);

        var child = parent.getChildNamed(title);
        child.QueryInterface(Components.interfaces.nsIMsgFolder);

        uri = child.URI;
    }

    return uri;
}

FeedService.prototype.__defineGetter__('fileHandler',
function() {
    return (
      Components
        .classes["@mozilla.org/network/io-service;1"]
          .getService(Components.interfaces.nsIIOService)
            .getProtocolHandler("file")
              .QueryInterface(Components.interfaces.nsIFileProtocolHandler)
    );
});

FeedService.prototype.__defineGetter__('appDirectoryService', function() {
    return (
      Components
        .classes["@mozilla.org/file/directory_service;1"]
          .getService(Components.interfaces.nsIProperties)
    );
});

FeedService.prototype.__defineGetter__('subscriptionsDS', function() {
    if ('_subscriptionsDS' in this)
        return this._subscriptionsDS;

    var url = this.fileHandler.getURLSpecFromFile(this.subscriptionsFile);

    this._subscriptionsDS = rdf.GetDataSource(url);
    if (!this._subscriptionsDS)
        throw("can't get subscriptions data source");

    // Note that it this point the datasource may not be loaded yet.
    // You have to QueryInterface it to nsIRDFRemoteDataSource and check
    // its "loaded" property to be sure.  You can also attach an observer
    // which will get notified when the load is complete.

    return this._subscriptionsDS;
});

FeedService.prototype.__defineGetter__('subscriptionsFile', function() {
    // Get the user's profile directory.
    var profileDir =
        this.appDirectoryService.get("ProfD", Components.interfaces.nsIFile);
    if (!profileDir)
        throw ("couldn't get the user's profile directory");
  
    // Get the user's subscriptions file.
    var file = profileDir.clone();
    file.append("feeds.rdf");

    // If the file doesn't exist, create it.
    if (!file.exists())
        this.createSubscriptionsFile(file);
  
    return file;
});

FeedService.prototype.createSubscriptionsFile = function(file) {
    var fileout = new LocalFile(file, MODE_WRONLY | MODE_CREATE);
    fileout.write('\
<?xml version="1.0"?>\n\
<RDF:RDF xmlns:dc="http://purl.org/dc/elements/1.1/"\n\
         xmlns:fz="' + FZ_NS + '"\n\
         xmlns:RDF="http://www.w3.org/1999/02/22-rdf-syntax-ns#">\n\
  <RDF:Description about="' + FZ_ROOT.Value + '">\n\
    <fz:feeds>\n\
      <RDF:Seq>\n\
      </RDF:Seq>\n\
    </fz:feeds>\n\
  </RDF:Description>\n\
</RDF:RDF>\n\
');
    fileout.close();
}

FeedService.prototype.__defineGetter__('itemsDS', function() {
    if ('_itemsDS' in this)
        return this._itemsDS;

    var url = this.fileHandler.getURLSpecFromFile(this.itemsFile);

    this._itemsDS = rdf.GetDataSource(url);
    if (!this._itemsDS)
        throw("can't get items data source");

    // Note that it this point the datasource may not be loaded yet.
    // You have to QueryInterface it to nsIRDFRemoteDataSource and check
    // its "loaded" property to be sure.  You can also attach an observer
    // which will get notified when the load is complete.

    return this._itemsDS;
});

FeedService.prototype.__defineGetter__('itemsFile', function() {
    // Get the user's profile directory.
    var profileDir =
        this.appDirectoryService.get("ProfD", Components.interfaces.nsIFile);
    if ( !profileDir )
        throw ("couldn't get the user's profile directory");
  
    // Get the user's subscriptions file.
    var file = profileDir.clone();
    file.append("feeditems.rdf");
  
    // If the file doesn't exist, create it.
    if (!file.exists()) {
        var fileout = new LocalFile(file, MODE_WRONLY | MODE_CREATE);
        fileout.write('\
<?xml version="1.0"?>\n\
<RDF:RDF xmlns:dc="http://purl.org/dc/elements/1.1/"\n\
         xmlns:fz="' + FZ_NS + '"\n\
         xmlns:RDF="http://www.w3.org/1999/02/22-rdf-syntax-ns#">\n\
</RDF:RDF>\n\
');
        fileout.close();
    }
    return file;
});

FeedService.prototype.__defineGetter__('subscriptionsList', function() {
    var list = this.subscriptionsDS.GetTarget(FZ_ROOT, FZ_FEEDS, true);
    //list = feeds.QueryInterface(Components.interfaces.nsIRDFContainer);
    list = list.QueryInterface(Components.interfaces.nsIRDFResource);
    list = rdfcontainer.MakeSeq(this.subscriptionsDS, list);
    return list;
});


/******************************************************************************/
// generic module registration code; adapted from nsUpdateService.js

var gModule = {
    _firstTime: true,
    _scriptsLoaded: false,
    
    _objects: {
        feedService: {
            CID: Components.ID("{5015eeed-7189-4ddc-984d-c87b5d948759}"),
            contractID: "@forumzilla.mozdev.org/feed/feedservice;1",
            className: "nsFeedService",
            factory: {
                createInstance: function (aOuter, aIID) {
                    if (aOuter != null)
                        throw Components.results.NS_ERROR_NO_AGGREGATION;
                    return new FeedService().QueryInterface(aIID);
                }
            }
        }
    },

    _loadScripts: function () {
        if (this._scriptsLoaded)
            return;

        var scriptLoader =
            Components
              .classes["@mozilla.org/moz/jssubscript-loader;1"]
                .getService(Components.interfaces.mozIJSSubScriptLoader);

        if (!scriptLoader)
            return;

        // Gives us LocalFile, an easy interface to files on the file system.
        scriptLoader.loadSubScript("chrome://forumzilla/content/file-utils.js");
        scriptLoader.loadSubScript("chrome://forumzilla/content/global.js");
        scriptLoader.loadSubScript("chrome://forumzilla/content/debug-utils.js");
        scriptLoader.loadSubScript("chrome://global/content/nsUserSettings.js");

        this._scriptsLoaded = true;
    },

    registerSelf: function (aComponentManager, aFileSpec, aLocation, aType) {
        if (this._firstTime) {
            this._firstTime = false;
            throw Components.results.NS_ERROR_FACTORY_REGISTER_AGAIN;
        }
        aComponentManager =
            aComponentManager
                .QueryInterface(Components.interfaces.nsIComponentRegistrar);
        
        for (var key in this._objects) {
            var obj = this._objects[key];
            aComponentManager.registerFactoryLocation(obj.CID,
                                                      obj.className,
                                                      obj.contractID,
                                                      aFileSpec,
                                                      aLocation,
                                                      aType);
        }
    },
    
    getClassObject: function (aComponentManager, aCID, aIID) {
        if (!aIID.equals(Components.interfaces.nsIFactory))
            throw Components.results.NS_ERROR_NOT_IMPLEMENTED;
        
        if (!this._scriptsLoaded)
            this._loadScripts();
        
        for (var key in this._objects) {
            if (aCID.equals(this._objects[key].CID))
              return this._objects[key].factory;
        }
        
        throw Components.results.NS_ERROR_NO_INTERFACE;
    },
  
    canUnload: function (aComponentManager) {
        return true;
    }
};

function NSGetModule(compMgr, fileSpec) {
    return gModule;
}
